package fr.mrdev023.vulkan_java.vk.utils;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

import org.lwjgl.system.MemoryStack;
import org.lwjgl.vulkan.VK10;

import fr.mrdev023.vulkan_java.vk.Instance;
import fr.mrdev023.vulkan_java.vk.PhysicalDevice;
import fr.mrdev023.vulkan_java.vk.VulkanError;
import fr.mrdev023.vulkan_java.vk.loggers.PhysicalDeviceCompatibilityLogger;
import fr.mrdev023.vulkan_java.vk.validators.PhysicalDeviceCompatibilityValidator;

public class SuitablePhysicalDeviceFinder {

    public static MatchResult findBestPhysicalDevice(Instance instance, PhysicalDeviceCompatibilityValidator validator)
            throws VulkanError {
        List<MatchResult> matchedPhysicalDevices = new ArrayList<>();

        try (MemoryStack stack = MemoryStack.stackPush()) {
            List<PhysicalDevice> physicalDevices = PhysicalDevice.getPhysicalDevices(instance, stack);
            for (PhysicalDevice physicalDevice : physicalDevices) {
                var matchResult = validator.validate(physicalDevice);
                if (matchResult.isSuitable) {
                    matchedPhysicalDevices.add(new MatchResult(physicalDevice, matchResult));
                }
                PhysicalDeviceCompatibilityLogger.log(physicalDevice, matchResult);
            }
        }

        return matchedPhysicalDevices.stream().min(Comparator.comparingInt(matchResult -> getScore(matchResult)))
                .orElse(null);
    }

    private static int getScore(MatchResult matchResult) {
        int deviceType = matchResult.physicalDevice.getVkPhysicalDeviceProperties().deviceType();

        return switch (deviceType) {
            case VK10.VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU -> 0;
            case VK10.VK_PHYSICAL_DEVICE_TYPE_INTEGRATED_GPU -> 1;
            case VK10.VK_PHYSICAL_DEVICE_TYPE_VIRTUAL_GPU -> 2;
            case VK10.VK_PHYSICAL_DEVICE_TYPE_CPU -> 3;
            case VK10.VK_PHYSICAL_DEVICE_TYPE_OTHER -> 4;
            default -> 5;
        };
    }

    public static class MatchResult {
        public final PhysicalDevice physicalDevice;
        public final PhysicalDeviceCompatibilityValidator.Result validatorResult;

        public MatchResult(PhysicalDevice physicalDevice, PhysicalDeviceCompatibilityValidator.Result validatorResult) {
            this.physicalDevice = physicalDevice;
            this.validatorResult = validatorResult;
        }
    }
}
